{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 58,
   "metadata": {},
   "outputs": [],
   "source": [
    "# This is a sample code for the algorithm introduced in the paper\n",
    "# 'Rollout-Based Charging Strategy for Electric Trucks with Hours-of-Service Regulations'\n",
    "# Copyright @ Yuchao Li, Ting Bai\n",
    "# Please cite the paper if the codes are used for academic publication\n",
    "\n",
    "# The solver used in the sample code for the linear programs is Gurobi\n",
    "# To check the available solvers in your computer, type 'print(cp.installed_solvers())'  \n",
    "# Further testing data, and variations of implementaion can be obtained upon request by writing to\n",
    "# tingbai@kth.se, number127578@gmail.com"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 59,
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "import cvxpy as cp\n",
    "import itertools\n",
    "import time\n",
    "from concurrent.futures import ProcessPoolExecutor, as_completed"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 60,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Parameters related to the road model\n",
    "class road_charge:\n",
    "    def __init__(self, tau, detour, Pcharging):\n",
    "        self.tau = tau              # Vector $[\\tau_0,\\dots,\\tau_N]$\n",
    "        self.detour = detour        # Vector $[d_0,\\dots,d_{N-1}]$\n",
    "        self.Pcharging = Pcharging  # Vector $[P_0,\\dots,P_{N-1}]$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 61,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Parameters related to the regulations and market\n",
    "class regulation_market:\n",
    "    def __init__(self, Tr, Td, Tbard, DeltaT, price):\n",
    "        self.Tr = Tr          # Minimum rest time, $T_r$\n",
    "        self.Td = Td          # Maximum continuous driving time, $T_d$\n",
    "        self.Tbard = Tbard    # Maximum total travel time per day, $\\bar{T}_d$\n",
    "        self.DeltaT = DeltaT  # Maximum deviation time related to the deadline, $\\Delta_T$\n",
    "        self.price = price    # Electricity price, $\\xi_k$ constant for all $k$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 62,
   "metadata": {},
   "outputs": [],
   "source": [
    "class EV_charge:\n",
    "    \n",
    "    def __init__(self, Pmax, eusable, eini, Pbar, tp):\n",
    "        self.Pmax = Pmax      # Maximum charging power of the truck, $P_{\\max}$\n",
    "        self.eusable = eusable# Usable energy of the truck with fully charged battery, $e_f-e_s$\n",
    "        self.e0 = eini        # Initial energy of the truck, $e_{ini}$\n",
    "        self.Pbar = Pbar      # Energy consumption per unit time, $\\bar{P}$\n",
    "        self.tp = tp          # Charging preparation time, $p_k$ constant for all $k$\n",
    "    #------------------------------------------------------------------------------\n",
    "    # Specify the charging problem given the parameters of the truck, road, and related reguations\n",
    "    def charge_problem(self,road_charge):\n",
    "        self.N = road_charge.detour.size\n",
    "        self.b_charging = np.zeros(self.N)          # Vector $[b_0,\\dots,b_{N-1}]$         \n",
    "        self.b_resting = np.zeros(self.N)           # Vector $[\\tilde{b}_0,\\dots,\\tilde{b}_{N-1}]$\n",
    "        # Vector $[\\bar{P}d_0,\\dots,\\bar{P}d_{N-1},0]$.\n",
    "        # The last element is one since the destination has reached.\n",
    "        self.esafe = np.append(self.Pbar*road_charge.detour.copy(),[0.]) #e_s\n",
    "        # Vector, elements are $\\min{P_k,P_{\\max}}$\n",
    "        self.Pactual = np.minimum(road_charge.Pcharging, self.Pmax)  \n",
    "    #------------------------------------------------------------------------------\n",
    "    # Timid base policy, setting $b_k=\\tilde{b}_k=1$ for all stations\n",
    "    def timid_policy(self):\n",
    "        self.timid_charging = np.ones(self.N)\n",
    "        self.timid_resting = np.ones(self.N)\n",
    "    #------------------------------------------------------------------------------\n",
    "    # Greedy base policy, setting $b_k=\\tilde{b}_k=1$ if the truck cannot reach station $k+1$\n",
    "    def greedy_policy(self,road_charge):\n",
    "        b = np.zeros(self.N)\n",
    "        e_current = self.e0\n",
    "        tau = road_charge.tau\n",
    "        detour = road_charge.detour\n",
    "\n",
    "        for i in range(self.N):\n",
    "            if (e_current-self.Pbar*(tau[i]+tau[i+1]))<self.esafe[i+1]:\n",
    "                b[i]=1\n",
    "                e_current = self.eusable-detour[i]*self.Pbar\n",
    "            else:\n",
    "                e_current -=self.Pbar*tau[i]\n",
    "  \n",
    "        self.greedy_charging = b.copy()\n",
    "        self.greedy_resting = b.copy()\n",
    "    #------------------------------------------------------------------------------\n",
    "    # Computing lower bound\n",
    "    def lower_bound(self, road_charge, regulation_market):\n",
    "        N = self.N\n",
    "        tau = road_charge.tau.copy()\n",
    "        detour = road_charge.detour.copy()\n",
    "\n",
    "        Pactual = self.Pactual\n",
    "        eusable = self.eusable\n",
    "        esafe = self.esafe.copy()\n",
    "        e0 = self.e0\n",
    "        Pbar = self.Pbar\n",
    "        tp = self.tp\n",
    "        Tr = regulation_market.Tr\n",
    "        Td = regulation_market.Td\n",
    "        Tbard = regulation_market.Tbard\n",
    "        DeltaT = regulation_market.DeltaT\n",
    "        Tbound = 1000000\n",
    "        price = regulation_market.price\n",
    "        epsilon = 0.001\n",
    "\n",
    "    \n",
    "        # if detour.dot(bbar) <= (Tbard-tau.sum())/2:\n",
    "        # This is the case where the total travel time constraint is fulfilled\n",
    "        t_charging = cp.Variable((N))                 # Vector, $[t_0,\\dots,t_{N-1}]$\n",
    "        t_extra = cp.Variable((N))                    # Vector, t_extra[i]+t_charging[i]>=Tr\n",
    "        c_continuous_driving = cp.Variable((N+1))     # Vector, $[c_0,\\dots,c_N]$\n",
    "        e_battery = cp.Variable((N+1))                # Vector, $[e_0,\\dots,e_N]$\n",
    "\n",
    "        delta_T = cp.Variable((N+1))\n",
    "\n",
    "        b_charging = cp.Variable((N))\n",
    "        b_resting = cp.Variable((N))\n",
    "        bbar = cp.Variable((N))\n",
    "        bhat = cp.Variable((N))\n",
    "        Delta_ehat = cp.Variable((N))\n",
    "        chat = cp.Variable((N))\n",
    "        t_ch_re = cp.Variable((N))\n",
    "\n",
    "\n",
    "        cost = 0\n",
    "        constr = []\n",
    "        constr += [c_continuous_driving[0] == tau[0]]\n",
    "        constr += [e_battery[0] == e0-Pbar*tau[0]]\n",
    "        constr += [delta_T[0] == 0]\n",
    "        for i in range(N):\n",
    "            constr += [b_charging[i]<=1,b_charging[i]>=0,b_resting[i]<=1,b_resting[i]>=0]\n",
    "            constr += [bhat[i]<=1,bhat[i]>=0,bbar[i]<=1,bbar[i]>=0]\n",
    "            constr += [bhat[i]<=b_charging[i],bhat[i]<=b_resting[i],bhat[i]>=b_charging[i]+b_resting[i]-1]\n",
    "            constr += [bbar[i]>=b_charging[i],bbar[i]>=b_resting[i],bbar[i]<=b_charging[i]+b_resting[i]]\n",
    "            constr += [Delta_ehat[i]<=b_charging[i]*Tbound,Delta_ehat[i]>=0]\n",
    "            constr += [t_charging[i]*Pactual[i]-Delta_ehat[i]<=(1-b_charging[i])*Tbound]\n",
    "            constr += [t_charging[i]*Pactual[i]-Delta_ehat[i]>=0]\n",
    "\n",
    "            constr += [chat[i]<=b_resting[i]*Tbound,chat[i]>=0]\n",
    "            constr += [c_continuous_driving[i]-chat[i]<=(1-b_resting[i])*Tbound]\n",
    "            constr += [c_continuous_driving[i]-chat[i]>=0]\n",
    "\n",
    "            constr += [t_ch_re[i]<=bhat[i]*Tbound,t_ch_re[i]>=0]\n",
    "            constr += [t_charging[i]-t_ch_re[i]<=(1-bhat[i])*Tbound]\n",
    "            constr += [t_charging[i]-t_ch_re[i]>=0]\n",
    "\n",
    "            constr += [c_continuous_driving[i]<= (Td-detour[i])]\n",
    "            constr += [c_continuous_driving[i+1] == tau[i+1]+bbar[i]*detour[i]+\\\n",
    "                 (c_continuous_driving[i]+bbar[i]*detour[i])-chat[i]-b_resting[i]*detour[i]]\n",
    "            constr += [e_battery[i] >= esafe[i]]\n",
    "            constr += [t_charging[i] <= (eusable-(e_battery[i]-Pbar*tau[i]))/Pactual[i]]\n",
    "            constr += [e_battery[i+1] == e_battery[i]+Delta_ehat[i]-2*bbar[i]*Pbar*detour[i]-Pbar*tau[i+1]]\n",
    "            constr += [t_extra[i]+Delta_ehat[i]/Pactual[i]+b_charging[i]*tp >= b_resting[i]*Tr]\n",
    "            constr += [Delta_ehat[i]/Pactual[i]+b_charging[i]*tp-t_ch_re[i]-bhat[i]*tp <= Tr-epsilon]\n",
    "            \n",
    "                \n",
    "\n",
    "            constr += [t_extra[i] >= 0, t_charging[i] >= 0]\n",
    "            constr += [t_extra[i] <= b_resting[i]*Tbound, t_charging[i] <= b_charging[i]*Tbound]\n",
    "            constr += [delta_T[i+1]==delta_T[i]+t_extra[i]+Delta_ehat[i]/Pactual[i]+b_charging[i]*tp]\n",
    "            cost += t_extra[i]+Delta_ehat[i]/Pactual[i]+b_charging[i]*tp+bbar[i]*2*detour[i]+Delta_ehat[i]/Pactual[i]*price\n",
    "\n",
    "        constr += [c_continuous_driving[N] <= Td, e_battery[N] >= esafe[N]]\n",
    "        constr += [delta_T[N] <= DeltaT, detour.T@bbar <= (Tbard-tau.sum())/2]\n",
    "\n",
    "\n",
    "        problem = cp.Problem(cp.Minimize(cost), constr)\n",
    "        lower_bd = problem.solve(solver='GUROBI')\n",
    "        lower_bd_t_charge = np.array(t_charging.value)\n",
    "        lower_bd_t_extra = np.array(t_extra.value)\n",
    "        lower_bd_b_charging = np.array(b_charging.value)\n",
    "        lower_bd_b_resting = np.array(b_resting.value)\n",
    "        return lower_bd, lower_bd_t_charge, lower_bd_t_extra, lower_bd_b_charging, lower_bd_b_resting\n",
    "    #------------------------------------------------------------------------------\n",
    "    # Compute relaxed base policy\n",
    "    def relaxed_policy(self, road_charge, regulation_market):\n",
    "        self.relaxed_ld, _, _, charging, resting = self.lower_bound(road_charge, regulation_market)\n",
    "        charging_tem = (charging>0)\n",
    "        self.relaxed_charging=charging_tem.astype(float)\n",
    "        resting_tem = (resting>0)\n",
    "        self.relaxed_resting=resting_tem.astype(float)\n",
    "    #------------------------------------------------------------------------------\n",
    "    # Compute timid, greedy, and relaxed base policies\n",
    "    def base_policies(self,road_charge,regulation_market):\n",
    "        self.charge_problem(road_charge)\n",
    "        self.timid_policy()\n",
    "        self.greedy_policy(road_charge)\n",
    "        self.relaxed_policy(road_charge, regulation_market)\n",
    "    #------------------------------------------------------------------------------\n",
    "    # Given a set of b_charging and b_resting, compute the optimal times stayed at each station\n",
    "    def time_schedule(self, b_charging, b_resting, road_charge, regulation_market):\n",
    "        N = self.N\n",
    "        tau = road_charge.tau.copy()\n",
    "        detour = road_charge.detour.copy()\n",
    "\n",
    "        Pactual = self.Pactual\n",
    "        eusable = self.eusable\n",
    "        esafe = self.esafe.copy()\n",
    "        e0 = self.e0\n",
    "        Pbar = self.Pbar\n",
    "        tp = self.tp\n",
    "        Tr = regulation_market.Tr\n",
    "        Td = regulation_market.Td\n",
    "        Tbard = regulation_market.Tbard\n",
    "        DeltaT = regulation_market.DeltaT\n",
    "        Tbound = 1000\n",
    "        price = regulation_market.price\n",
    "        epsilon = 0.001\n",
    "\n",
    "        # bbar is the vector of $[\\bar{b}_0,\\dots,\\bar{b}_{N-1}]$\n",
    "        bbar = np.logical_or(b_charging,b_resting).astype(float)\n",
    "\n",
    "        if detour.dot(bbar) <= (Tbard-tau.sum())/2:\n",
    "            \n",
    "            # This is the case where the total travel time constraint is fulfilled\n",
    "            t_charging = cp.Variable((N))                 # Vector, $[t_0,\\dots,t_{N-1}]$\n",
    "            delta_e = cp.Variable((N))                    # Vector, $[\\Delta e_0,\\dots,\\Delta e_{N-1}]$\n",
    "            t_extra = cp.Variable((N))                    # Vector, t_extra[i]+t_charging[i]>=Tr\n",
    "            c_continuous_driving = cp.Variable((N+1))     # Vector, $[c_0,\\dots,c_N]$\n",
    "            e_battery = cp.Variable((N+1))                # Vector, $[e_0,\\dots,e_N]$\n",
    "\n",
    "            delta_T = cp.Variable((N+1))\n",
    "\n",
    "            cost = 0\n",
    "            constr = []\n",
    "            constr += [c_continuous_driving[0] == tau[0]]\n",
    "            constr += [e_battery[0] == e0-Pbar*tau[0]]\n",
    "            constr += [delta_T[0] == 0]\n",
    "            for i in range(N):\n",
    "                constr += [c_continuous_driving[i]<= (Td-detour[i])]\n",
    "                constr += [c_continuous_driving[i+1] == tau[i+1]+bbar[i]*detour[i]+(1-b_resting[i])*(c_continuous_driving[i]+bbar[i]*detour[i])]\n",
    "                constr += [e_battery[i] >= esafe[i]]\n",
    "                constr += [delta_e[i]+e_battery[i] <= eusable+Pbar*tau[i]]\n",
    "                constr += [delta_e[i] == t_charging[i]*Pactual[i]]\n",
    "                constr += [e_battery[i+1] == e_battery[i]+b_charging[i]*delta_e[i]-2*bbar[i]*Pbar*detour[i]-Pbar*tau[i+1]]\n",
    "                constr += [t_extra[i]+b_charging[i]*(t_charging[i]+tp) >= b_resting[i]*Tr]\n",
    "                constr += [(1-b_resting[i])*b_charging[i]*(t_charging[i]+tp) <= Tr-epsilon]\n",
    "                constr += [t_extra[i] >= 0, t_charging[i] >= 0]\n",
    "                constr += [t_extra[i] <= b_resting[i]*Tbound, t_charging[i] <= b_charging[i]*Tbound]\n",
    "                constr += [delta_T[i+1]==delta_T[i]+t_extra[i]+b_charging[i]*(t_charging[i]+tp)]\n",
    "                cost += t_extra[i]+b_charging[i]*(t_charging[i]+tp)+bbar[i]*2*detour[i]+b_charging[i]*t_charging[i]*price\n",
    "\n",
    "            constr += [c_continuous_driving[N] <= Td, e_battery[N] >= esafe[N]]\n",
    "            constr += [delta_T[N] <= DeltaT]\n",
    "\n",
    "            problem = cp.Problem(cp.Minimize(cost), constr)\n",
    "            op_value = problem.solve(solver='GUROBI')\n",
    "            op_t_charge = np.array(t_charging.value)\n",
    "            op_t_extra = np.array(t_extra.value)\n",
    "        else:\n",
    "            op_value = np.inf\n",
    "            op_t_charge = np.nan*np.ones(N)\n",
    "            op_t_extra = np.nan*np.ones(N)\n",
    "            \n",
    "        return op_value, op_t_charge, op_t_extra # upper bound\n",
    "    #------------------------------------------------------------------------------\n",
    "    # Rollout algorithm for solving the charging problem\n",
    "    def rollout(self, b_charging, b_resting, road_charge, regulation_market):\n",
    "        rollout = np.inf\n",
    "        temp = np.inf\n",
    "\n",
    "        rollout_charging = b_charging.copy()\n",
    "        rollout_resting = b_resting.copy()\n",
    "        rollout_t_charge = cp.Variable(self.N) \n",
    "        rollout_t_extra = cp.Variable(self.N) \n",
    "    \n",
    "        for i in reversed(range(self.N)):\n",
    "            b_charging = rollout_charging.copy()\n",
    "            b_resting = rollout_resting.copy()\n",
    "            for j in range(4):\n",
    "                if j == 0:\n",
    "                    b_charging[i] = 0\n",
    "                    b_resting[i] = 0\n",
    "                elif j == 1:\n",
    "                    b_charging[i] = 0\n",
    "                    b_resting[i] = 1\n",
    "                elif j == 2:\n",
    "                    b_charging[i] = 1\n",
    "                    b_resting[i] = 0\n",
    "                else:\n",
    "                    b_charging[i] = 1\n",
    "                    b_resting[i] = 1\n",
    "                temp, op_t_charge, op_t_extra = self.time_schedule(b_charging, b_resting, road_charge, regulation_market)\n",
    "                if temp is not None:\n",
    "                    if rollout > temp:\n",
    "                        rollout_charging = b_charging.copy()\n",
    "                        rollout_resting = b_resting.copy()\n",
    "                        rollout = temp\n",
    "                        rollout_t_charge = op_t_charge\n",
    "                        rollout_t_extra = op_t_extra\n",
    "                \n",
    "    \n",
    "        return rollout, rollout_t_charge, rollout_t_extra, rollout_charging, rollout_resting\n",
    "\n",
    "    #------------------------------------------------------------------------------\n",
    "    # Parametrized implementation of the rollout algorithm\n",
    "    def rollout_para(self, b_charging_ini, b_resting_ini, road_charge, regulation_market):\n",
    "        rollout = np.inf\n",
    "        temp = np.inf\n",
    "\n",
    "        rollout_charging = b_charging_ini.copy()\n",
    "        rollout_resting = b_resting_ini.copy()\n",
    "        rollout_t_charge = rollout_charging.copy()\n",
    "        rollout_t_extra=rollout_charging.copy()\n",
    "        \n",
    "        N = self.N\n",
    "        tau = road_charge.tau.copy()\n",
    "        detour = road_charge.detour.copy()\n",
    "\n",
    "        Pactual = self.Pactual\n",
    "        eusable = self.eusable\n",
    "        esafe = self.esafe.copy()\n",
    "        e0 = self.e0\n",
    "        Pbar = self.Pbar\n",
    "        tp = self.tp\n",
    "        Tr = regulation_market.Tr\n",
    "        Td = regulation_market.Td\n",
    "        Tbard = regulation_market.Tbard\n",
    "        DeltaT = regulation_market.DeltaT\n",
    "        Tbound = 1000\n",
    "        price = regulation_market.price\n",
    "        epsilon = 0.001\n",
    "        \n",
    "        b_charging = cp.Parameter(N,nonneg=True)\n",
    "        b_resting = cp.Parameter(N,nonneg=True)\n",
    "        bbar = cp.Parameter(N,nonneg=True)\n",
    "        bhat = cp.Parameter(N,nonneg=True)\n",
    "            \n",
    "            \n",
    "        t_charging = cp.Variable((N))                 # Vector, $[t_0,\\dots,t_{N-1}]$\n",
    "        delta_e = cp.Variable((N))                    # Vector, $[\\Delta e_0,\\dots,\\Delta e_{N-1}]$\n",
    "        t_extra = cp.Variable((N))                    # Vector, t_extra[i]+t_charging[i]>=Tr\n",
    "        c_continuous_driving = cp.Variable((N+1))     # Vector, $[c_0,\\dots,c_N]$\n",
    "        e_battery = cp.Variable((N+1))                # Vector, $[e_0,\\dots,e_N]$\n",
    "\n",
    "        delta_T = cp.Variable((N+1))\n",
    "\n",
    "        cost = 0\n",
    "        constr = []\n",
    "        constr += [c_continuous_driving[0] == tau[0]]\n",
    "        constr += [e_battery[0] == e0-Pbar*tau[0]]\n",
    "        constr += [delta_T[0] == 0]\n",
    "        for i in range(N):\n",
    "            constr += [c_continuous_driving[i]<= (Td-detour[i])]\n",
    "            constr += [c_continuous_driving[i+1] == tau[i+1]+bbar[i]*detour[i]+\\\n",
    "                      (1-b_resting[i])*c_continuous_driving[i]+bbar[i]*detour[i]-b_resting[i]*detour[i]]\n",
    "            constr += [e_battery[i] >= esafe[i]]\n",
    "            constr += [delta_e[i]+e_battery[i] <= eusable+Pbar*tau[i]]\n",
    "            constr += [delta_e[i] == t_charging[i]*Pactual[i]]\n",
    "            constr += [e_battery[i+1] == e_battery[i]+b_charging[i]*delta_e[i]-2*bbar[i]*Pbar*detour[i]-Pbar*tau[i+1]]\n",
    "            constr += [t_extra[i]+b_charging[i]*(t_charging[i]+tp) >= b_resting[i]*Tr]\n",
    "            constr += [(1-b_resting[i])*b_charging[i]*(t_charging[i]+tp) <= Tr-epsilon]\n",
    "            constr += [t_extra[i] >= 0, t_charging[i] >= 0]\n",
    "            constr += [t_extra[i] <= b_resting[i]*Tbound, t_charging[i] <= b_charging[i]*Tbound]\n",
    "            constr += [delta_T[i+1]==delta_T[i]+t_extra[i]+b_charging[i]*(t_charging[i]+tp)]\n",
    "            cost += t_extra[i]+b_charging[i]*(t_charging[i]+tp)+bbar[i]*2*detour[i]+b_charging[i]*t_charging[i]*price\n",
    "\n",
    "        constr += [c_continuous_driving[N] <= Td, e_battery[N] >= esafe[N]]\n",
    "        constr += [delta_T[N] <= DeltaT, detour.T@bbar <= (Tbard-tau.sum())/2]\n",
    "        \n",
    "        prob = cp.Problem(cp.Minimize(cost), constr)\n",
    "        \n",
    "        comb = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])\n",
    "\n",
    "         \n",
    "        for i in range(self.N):\n",
    "            \n",
    "            for j in range(4):\n",
    "                b_charging_tem = rollout_charging.copy()\n",
    "                b_resting_tem = rollout_resting.copy()\n",
    "\n",
    "                b_charging_tem[i] = comb[j, 0]\n",
    "                b_resting_tem[i] = comb[j, 1]\n",
    "\n",
    "                b_charging.value = b_charging_tem.copy()\n",
    "                b_resting.value = b_resting_tem.copy()\n",
    "                bbar.value = np.logical_or(b_charging_tem.copy(), b_resting_tem.copy()).astype(float)\n",
    "                bhat.value = np.multiply(b_charging_tem, b_resting_tem)\n",
    "\n",
    "                temp = prob.solve(solver='GUROBI')\n",
    "                if temp is not None:\n",
    "                    if rollout > temp:\n",
    "                        rollout_charging = b_charging_tem.copy()\n",
    "                        rollout_resting = b_resting_tem.copy()\n",
    "                        rollout = temp\n",
    "                        rollout_t_charge = np.array(t_charging.value)\n",
    "                        rollout_t_extra = np.array(t_extra.value)\n",
    "                \n",
    "        \n",
    "        return rollout, rollout_charging, rollout_resting, rollout_t_charge, rollout_t_extra\n",
    "    \n",
    "    # Rollout for earch combination\n",
    "    def rollout_per_combination(self, i, b_charging, b_resting, road_charge, regulation_market):\n",
    "        temp = np.inf\n",
    "        for j in range(4):\n",
    "            if j == 0:\n",
    "                b_charging[i] = 0\n",
    "                b_resting[i] = 0\n",
    "            elif j == 1:\n",
    "                b_charging[i] = 0\n",
    "                b_resting[i] = 1\n",
    "            elif j == 2:\n",
    "                b_charging[i] = 1\n",
    "                b_resting[i] = 0\n",
    "            else:\n",
    "                b_charging[i] = 1\n",
    "                b_resting[i] = 1\n",
    "            temp, op_t_charge, op_t_extra = self.time_schedule(b_charging, b_resting, road_charge, regulation_market)\n",
    "            if temp < np.inf:\n",
    "                return temp, b_charging.copy(), b_resting.copy(), op_t_charge, op_t_extra\n",
    "        return np.inf, b_charging.copy(), b_resting.copy(), None, None\n",
    "    \n",
    "    # Parallel implementation of rollout\n",
    "    def rollout_parallel(self, b_charging, b_resting, road_charge, regulation_market):\n",
    "        rollout = np.inf\n",
    "        rollout_charging = b_charging.copy()\n",
    "        rollout_resting = b_resting.copy()\n",
    "        rollout_t_charge = b_charging.copy()\n",
    "        rollout_t_extra = b_charging.copy()\n",
    "\n",
    "        with ProcessPoolExecutor() as executor:\n",
    "            futures = {executor.submit(self.rollout_per_combination, i, b_charging.copy(), b_resting.copy(), road_charge, regulation_market) \n",
    "                       for i in reversed(range(self.N))}\n",
    "            for future in as_completed(futures):\n",
    "                temp, temp_charging, temp_resting, temp_t_charge, temp_t_extra = future.result()\n",
    "                if temp < rollout:\n",
    "                    rollout = temp\n",
    "                    rollout_charging = temp_charging\n",
    "                    rollout_resting = temp_resting\n",
    "                    rollout_t_charge = temp_t_charge\n",
    "                    rollout_t_extra = temp_t_extra\n",
    "\n",
    "        return rollout, rollout_t_charge, rollout_t_extra, rollout_charging, rollout_resting\n",
    "        \n",
    "    \n",
    "    #------------------------------------------------------------------------------\n",
    "    # Computing the best charging solution by enumerating all the combinations of the binary variables\n",
    "    def best_exhaust(self, road_charge, regulation_market):\n",
    "        N = self.N\n",
    "        best=np.inf\n",
    "        current=np.inf\n",
    "        best_charging=np.zeros(N)\n",
    "        best_resting=np.zeros(N)\n",
    "\n",
    "        vector=[0,1]\n",
    "        b_charge_combinations=list(itertools.product(vector,repeat=N))# all the combinations of the binary variables for b_charging\n",
    "        b_rest_combinations=list(itertools.product(vector,repeat=N))\n",
    "        for b_1 in b_charge_combinations:\n",
    "            for b_2 in b_rest_combinations:\n",
    "                current,t_charge,t_extra=self.time_schedule(np.array(b_1), np.array(b_2),road_charge, regulation_market)\n",
    "                if current is not None:\n",
    "                    if best > current:\n",
    "                        best_charging = np.array(b_1).copy()\n",
    "                        best_resting = np.array(b_2).copy()\n",
    "                        best_t_charge=t_charge.copy()\n",
    "                        best_t_extra=t_extra.copy()\n",
    "                        best = current.copy()\n",
    "                 \n",
    "\n",
    "        self.best_charging = best_charging.copy()\n",
    "        self.best_resting = best_resting.copy()\n",
    "        self.best_t_charge = best_t_charge.copy()\n",
    "        self.best_t_extra = best_t_extra.copy()\n",
    "        self.best = best\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Testing data used in the paper\n",
    "#N=5\n",
    "#{'tau': [23.4219, 68.9947, 24.1869, 107.553, 89.3284, 24.4093],\n",
    "# 'detour': [11.8715, 13.0527, 10.1167, 15.2695, 17.0627],\n",
    "# 'DeltaT': 150}\n",
    "#N=6\n",
    "#{'tau': [24.7328, 17.3314, 52.2553, 24.4523, 76.4669, 35.5783, 79.6548],\n",
    "# 'detour': [25.7005, 4.8874, 4.513, 20.3038, 6.0249, 3.2189],\n",
    "# 'DeltaT': 150}\n",
    "#N=7\n",
    "#{'tau': [30.1441, 39.1227, 54.9621, 70.438, 15.7162, 73.1713, 62.0197, 26.0694],\n",
    "# 'detour': [9.8573, 7.7971, 3.6675, 0.5792, 1.8247, 6.0237, 7.4384],\n",
    "# 'DeltaT': 150}\n",
    "#N=8\n",
    "#{'tau': [36.4262, 70.3561, 15.5605, 73.1098, 78.214, 22.4032, 53.2017, 17.8092, 25.7993],\n",
    "# 'detour': [2.2147, 2.2827, 1.8254, 6.023, 19.2596, 4.2088, 10.57, 24.568],\n",
    "# 'DeltaT': 150}\n",
    "#N=9\n",
    "#{'tau': [29.6962, 37.8472, 102.0682, 73.6421, 17.1241, 69.6287, 54.3168, 23.1439, 31.5338, 41.805, 43.3317],\n",
    "# 'detour': [3.4023, 0.7798, 6.0249, 1.8252, 2.3276, 3.6682, 0.1866, 3.7495, 6.1948],\n",
    "# 'DeltaT': 220}\n",
    "#N=10\n",
    "#{'tau': [30.1441, 39.1227, 54.9621, 70.438, 15.7162, 73.1713, 78.1987, 22.4278, 66.7064, 55.2526, 28.2965],\n",
    "# 'detour': [9.8573, 7.7971, 3.6675, 0.5792, 1.8247, 6.0237, 19.2602, 4.2078, 0.305, 11.0378],\n",
    "# 'DeltaT': 220}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 63,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Example run for the code\n",
    "\n",
    "tau_list=[23.4219, 68.9947, 24.1869, 107.553, 89.3284, 24.4093]\n",
    "\n",
    "detour_list=[11.8715, 13.0527, 10.1167, 15.2695, 17.0627]\n",
    "# compute the lower bound of the initial battery\n",
    "tau_0=tau_list[0]\n",
    "detour_0=detour_list[0]\n",
    "        \n",
    "# Parameter setting\n",
    "tau=np.array(tau_list)\n",
    "detour=np.array(detour_list)\n",
    "Pcharging = np.array([300, 300, 300, 300, 300])/60\n",
    "Pmax = 375/60\n",
    "Pbar = 110/60\n",
    "tp = 6\n",
    "Tr = 45\n",
    "Td = 4.5*60\n",
    "Tbard = 9*60\n",
    "DeltaT = 2.5*60\n",
    "price = 0.36/0.4\n",
    "Pbar=110/60\n",
    "eusable = 468\n",
    "eini = 0.8*eusable"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 64,
   "metadata": {},
   "outputs": [],
   "source": [
    "road = road_charge(tau, detour, Pcharging)\n",
    "regulation = regulation_market(Tr, Td, Tbard, DeltaT, price)\n",
    "ev = EV_charge(Pmax, eusable, eini, Pbar, tp)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 65,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[21.76441667 23.92995    18.54728333 27.99408333 31.28161667  0.        ] [0. 0. 1. 0. 0.]\n"
     ]
    },
    {
     "data": {
      "text/plain": [
       "(inf, array(None, dtype=object), array(None, dtype=object))"
      ]
     },
     "execution_count": 65,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "ev.base_policies(road,regulation)\n",
    "print(ev.esafe,ev.greedy_charging)\n",
    "ev.time_schedule(ev.timid_charging,ev.timid_resting,road,regulation)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 66,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1.661758553999789\n"
     ]
    }
   ],
   "source": [
    "# Test parametrized implementation of rollout\n",
    "pre = time.perf_counter()\n",
    "ev.rollout_para(ev.relaxed_charging,ev.relaxed_resting, road, regulation)\n",
    "print(time.perf_counter()-pre)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 67,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "0.3613530269999501\n"
     ]
    }
   ],
   "source": [
    "# Test parallel implementation of rollout\n",
    "pre = time.perf_counter()\n",
    "ev.rollout_parallel(ev.relaxed_charging,ev.relaxed_resting, road, regulation)\n",
    "print(time.perf_counter()-pre)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 68,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "69.24578801500002\n"
     ]
    }
   ],
   "source": [
    "pre = time.perf_counter()\n",
    "ev.best_exhaust(road, regulation)\n",
    "print(time.perf_counter()-pre)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
